"""
Clase base universal para patrones x.x.x.x - TIPO REGISTRADO
"""
from class_base import ClassBase
import sympy
import re

class FourBytes(ClassBase):
    """
    Representación universal de patrones de 4 elementos
    con manejo interno de símbolos y conversión perezosa a SymPy
    """
    
    def __init__(self, dotted_string):
        self.original = str(dotted_string)
        self.elements = self.original.split('.')
        
        if len(self.elements) != 4:
            raise ValueError(f"FourBytes requiere exactamente 4 elementos: {dotted_string}")
        
        self.has_symbols = any(not elem.isdigit() for elem in self.elements)
        
        if self.has_symbols:
            # Modo algebraico: mantener símbolos INTERNAMENTE, no convertir a SymPy aún
            self._symbols = self._extract_symbols()
            self._numeric_value = None  # No hay valor numérico directo
            super().__init__(self.original, self.original)  # ClassBase, NO SympyClassBase
        else:
            # Modo numérico: validar rangos y convertir
            self._numeric_elements = []
            for elem in self.elements:
                try:
                    num = int(elem)
                    if not (0 <= num <= 255):
                        raise ValueError(f"Elemento fuera de rango [0-255]: {num}")
                    self._numeric_elements.append(num)
                except ValueError:
                    raise ValueError(f"Elemento inválido: '{elem}'")
            
            # Crear valor como entero de 32 bits (para IPs)
            self._numeric_value = (self._numeric_elements[0] << 24 | 
                                 self._numeric_elements[1] << 16 | 
                                 self._numeric_elements[2] << 8 | 
                                 self._numeric_elements[3])
            self._symbols = []
            super().__init__(self._numeric_value, self.original)  # ClassBase
    
    def _extract_symbols(self):
        """Extrae símbolos únicos de todos los elementos"""
        symbols = []
        for elem in self.elements:
            if not elem.isdigit():
                if elem not in symbols:
                    symbols.append(elem)
        return symbols
    
    def substitute(self, **kwargs):
        """Sustitución NATIVA - mantiene como FourBytes"""
        if not self.has_symbols:
            return self  # Sin símbolos, retornar el mismo objeto
        
        new_elements = []
        for elem in self.elements:
            if elem in kwargs:
                new_elements.append(str(kwargs[elem]))
            else:
                new_elements.append(elem)
        
        # Retornar nuevo FourBytes con valores sustituidos
        return FourBytes('.'.join(new_elements))
    
    def to_sympy(self):
        """CONVERSIÓN EXPLÍCITA a SymPy cuando se necesite álgebra compleja"""
        if self.has_symbols:
            # Crear expresión escalar para compatibilidad con IP4 (no Matrix)
            symbolic_elements = []
            for elem in self.elements:
                if elem.isdigit():
                    symbolic_elements.append(int(elem))
                else:
                    symbolic_elements.append(sympy.Symbol(elem))
            
            # Crear expresión como entero de 32 bits simbólico (compatible con operaciones IP)
            result = (symbolic_elements[0] * 2**24 + 
                     symbolic_elements[1] * 2**16 + 
                     symbolic_elements[2] * 2**8 + 
                     symbolic_elements[3])
            
            return result
        else:
            # Sin símbolos, retornar entero SymPy simple
            return sympy.sympify(self._numeric_value)
    
    def is_valid_ip_range(self):
        """Verifica si todos los elementos están en rango IP válido"""
        if self.has_symbols:
            return None  # No se puede verificar con símbolos
        return all(0 <= elem <= 255 for elem in self._numeric_elements)
    
    def __getitem__(self, index):
        """Acceso a elementos individuales"""
        if self.has_symbols:
            elem = self.elements[index]
            if elem.isdigit():
                return int(elem)
            else:
                return sympy.Symbol(elem)
        else:
            return self._numeric_elements[index]
    
    def to_ip_int(self):
        """Convierte a entero de 32 bits (para IPs) - MANTIENE NATIVO"""
        if self.has_symbols:
            # Retornar expresión simbólica escalar, NO convertir a SymPy aún
            return f"ip_int({self.original})"
        else:
            return self._numeric_value
    
    @staticmethod
    def _int_to_fourbytes(value):
        """Convierte entero de 32 bits a formato dotted"""
        return f"{(value >> 24) & 0xFF}.{(value >> 16) & 0xFF}.{(value >> 8) & 0xFF}.{value & 0xFF}"
    
    # ========== OPERADORES ARITMÉTICOS NATIVOS (mantienen FourBytes) ==========
    
    def __add__(self, other):
        """Suma nativa - mantiene como FourBytes cuando es posible"""
        if self.has_symbols:
            # Con símbolos, crear expresión simbólica pero mantener FourBytes
            if isinstance(other, FourBytes):
                if other.has_symbols:
                    new_expr = f"({self.original}) + ({other.original})"
                else:
                    new_expr = f"({self.original}) + {other._numeric_value}"
            elif isinstance(other, (int, float)):
                new_expr = f"({self.original}) + {other}"
            else:
                # Para otros tipos, convertir a SymPy
                return self.to_sympy() + other
            
            return FourBytes(new_expr)
        
        # Sin símbolos: aritmética numérica directa
        if isinstance(other, FourBytes):
            if other.has_symbols:
                # El otro tiene símbolos, crear expresión
                new_expr = f"{self._numeric_value} + ({other.original})"
                return FourBytes(new_expr)
            else:
                # Ambos numéricos
                result_value = (self._numeric_value + other._numeric_value) & 0xFFFFFFFF
        elif isinstance(other, (int, float)):
            result_value = (self._numeric_value + int(other)) & 0xFFFFFFFF
        else:
            # Para tipos complejos, convertir a SymPy
            return self.to_sympy() + other
        
        # Convertir resultado de vuelta a formato dotted
        result_str = self._int_to_fourbytes(result_value)
        return FourBytes(result_str)
    
    def __sub__(self, other):
        """Resta nativa"""
        if self.has_symbols:
            if isinstance(other, FourBytes):
                if other.has_symbols:
                    new_expr = f"({self.original}) - ({other.original})"
                else:
                    new_expr = f"({self.original}) - {other._numeric_value}"
            elif isinstance(other, (int, float)):
                new_expr = f"({self.original}) - {other}"
            else:
                return self.to_sympy() - other
            
            return FourBytes(new_expr)
        
        # Sin símbolos
        if isinstance(other, FourBytes):
            if other.has_symbols:
                new_expr = f"{self._numeric_value} - ({other.original})"
                return FourBytes(new_expr)
            else:
                result_value = (self._numeric_value - other._numeric_value) & 0xFFFFFFFF
        elif isinstance(other, (int, float)):
            result_value = (self._numeric_value - int(other)) & 0xFFFFFFFF
        else:
            return self.to_sympy() - other
        
        result_str = self._int_to_fourbytes(result_value)
        return FourBytes(result_str)
    
    def ToBase(self, base):
        """Convierte cada elemento a la base especificada"""
        # Necesitamos importar IntBase desde el registro
        try:
            from type_registry import get_registered_base_context
            context = get_registered_base_context()
            IntBase = context.get('IntBase')
            if not IntBase:
                raise ImportError("IntBase no disponible en el registro")
        except ImportError:
            # Fallback: usar el método directo 
            def _convert_to_base_string(value, base):
                if base == 2:
                    return bin(value)[2:]
                elif base == 8:
                    return oct(value)[2:]
                elif base == 16:
                    return hex(value)[2:].upper()
                elif base == 10:
                    return str(value)
                else:
                    digits = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                    result = ""
                    while value > 0:
                        result = digits[value % base] + result
                        value //= base
                    return result or "0"
            
            if self.has_symbols:
                # Con símbolos, mantener estructura simbólica
                base_elements = []
                for elem in self.elements:
                    if elem.isdigit():
                        num = int(elem)
                        base_str = _convert_to_base_string(num, base)
                        base_elements.append(f"{base}#{base_str}")
                    else:
                        base_elements.append(f"{base}#{elem}")
                return '.'.join(base_elements)
            else:
                # Sin símbolos, conversión directa
                base_elements = []
                for num in self._numeric_elements:
                    base_str = _convert_to_base_string(num, base)
                    base_elements.append(f"{base}#{base_str}")
                return '.'.join(base_elements)
        
        # Versión con IntBase disponible
        if self.has_symbols:
            base_elements = []
            for elem in self.elements:
                if elem.isdigit():
                    num = int(elem)
                    base_str = IntBase._convert_to_base_string(num, base)
                    base_elements.append(f"{base}#{base_str}")
                else:
                    base_elements.append(f"{base}#{elem}")
            return '.'.join(base_elements)
        else:
            base_elements = []
            for num in self._numeric_elements:
                base_str = IntBase._convert_to_base_string(num, base)
                base_elements.append(f"{base}#{base_str}")
            return '.'.join(base_elements)
    
    def ToCIDR(self, prefix_length):
        """Convierte a notación CIDR"""
        return f"{self.original}/{prefix_length}"
    
    def ToHex(self):
        """Conversión a hexadecimal"""
        return self.ToBase(16)
    
    def ToBinary(self):
        """Conversión a binario"""
        return self.ToBase(2)
    
    def ToOctal(self):
        """Conversión a octal"""
        return self.ToBase(8)
    
    @property
    def value(self):
        """Propiedad de acceso al valor - mantiene nativo"""
        if self.has_symbols:
            return self.original  # Mantener como string simbólico
        else:
            return self._numeric_value  # Valor numérico directo
    
    def __str__(self):
        """Representación como cadena"""
        return self.original
    
    def __repr__(self):
        """Representación para debug"""
        return f"FourBytes('{self.original}')"
    
    # ========== MÉTODOS PARA COMPATIBILIDAD CON SYMPY ==========
    
    def _sympystr(self, printer):
        """Controla cómo SymPy representa este objeto"""
        # Mantener representación nativa, no convertir a número
        return f"FourBytes('{self.original}')"
    
    def __radd__(self, other):
        """Suma reversa - cuando el otro operando es el primero"""
        if isinstance(other, (int, float)):
            return self.__add__(other)
        else:
            # Para otros tipos, dejar que SymPy maneje
            return NotImplemented
    
    def __rsub__(self, other):
        """Resta reversa"""
        if isinstance(other, (int, float)):
            # other - self
            if self.has_symbols:
                new_expr = f"{other} - ({self.original})"
                return FourBytes(new_expr)
            else:
                result_value = (int(other) - self._numeric_value) & 0xFFFFFFFF
                result_str = self._int_to_fourbytes(result_value)
                return FourBytes(result_str)
        else:
            return NotImplemented
    
    @staticmethod
    def Helper(input_str):
        """Ayuda contextual para FourBytes"""
        if re.search(r'\d+\.\d+\.\d+\.\d+', input_str):
            return '''FourBytes - Patrones x.x.x.x con álgebra simbólica
            
Sintaxis: x.y.z.w (donde cada elemento puede ser número o símbolo)
Ejemplos:
  192.168.1.1      → FourBytes('192.168.1.1') 
  255.255.0.0      → FourBytes('255.255.0.0')
  10.x.1.y         → FourBytes('10.x.1.y') = simbólico
  
Aritmética de 32-bit:
  192.168.1.1 + 5  → FourBytes('192.168.1.6')
  10.0.0.0 + 256   → FourBytes('10.0.1.0')
  
Métodos disponibles:
  .ToBase(base)       - Convierte cada elemento a base específica
  .ToHex()           - Convierte a hexadecimal  
  .ToBinary()        - Convierte a binario
  .substitute(x=val) - Sustituye símbolos
  .to_sympy()        - Conversión explícita a SymPy'''
        return None
    
    @staticmethod
    def PopupFunctionList():
        """Lista de métodos para autocompletado"""
        return [
            ("ToBase", "Convierte cada elemento a base específica"),
            ("ToHex", "Convierte cada elemento a hexadecimal"),
            ("ToBinary", "Convierte cada elemento a binario"),
            ("ToOctal", "Convierte cada elemento a octal"),
            ("ToCIDR", "Añade notación CIDR"),
            ("substitute", "Sustituye símbolos por valores"),
            ("to_sympy", "Conversión explícita a SymPy para álgebra"),
            ("is_valid_ip_range", "Verifica si es rango IP válido"),
        ]
    
    @staticmethod
    def get_tokenization_patterns():
        """
        Define cómo esta clase debe ser reconocida y tokenizada
        
        Returns:
            List[Dict]: Lista de reglas de tokenización
        """
        return [
            {
                'pattern': r'(?<!FourBytes\(")(?<!")([a-zA-Z0-9_]+\.[a-zA-Z0-9_]+\.[a-zA-Z0-9_]+\.[a-zA-Z0-9_]+)(?!")',
                'replacement': lambda match: f'FourBytes("{match.group(1)}")',
                'priority': 10,  # MENOR prioridad que IntBase (patrón más general)
                'description': 'Patrones de 4 elementos: 192.168.1.1, 10.x.y.z'
            }
        ]

def register_classes_in_module():
    """Registro de FourBytes en el sistema de tipos"""
    return [
        ("FourBytes", FourBytes, "ClassBase", {
            "add_lowercase": True,
            "supports_brackets": False,  # Se maneja por tokenización
            "is_fundamental": True,      # Clase fundamental del sistema
            "description": "Patrones universales x.x.x.x con álgebra simbólica"
        }),
    ] 