"""
Clase híbrida para conversión LaTeX - ADAPTADA AL NUEVO SISTEMA
"""
from sympy_Base import SympyClassBase
import sympy
import re


class Class_LaTeX(SympyClassBase):
    """Clase híbrida para conversión de expresiones a formato LaTeX"""
    
    def __new__(cls, value_input):
        """Crear objeto SymPy válido"""
        obj = SympyClassBase.__new__(cls)
        return obj
    
    def __init__(self, expression):
        """Inicialización de LaTeX"""
        # Si el input es string, intentar parsearlo como expresión de SymPy
        if isinstance(expression, str):
            try:
                # Parsear la expresión string a objeto SymPy
                parsed_expr = sympy.sympify(expression)
                self._expression = parsed_expr
            except Exception as e:
                # Si no se puede parsear, mantener como string
                self._expression = expression
        else:
            # Si ya es un objeto SymPy, usarlo directamente
            self._expression = expression
        
        # Generar código LaTeX
        try:
            if hasattr(self._expression, '__class__') and hasattr(sympy, 'latex'):
                self._latex_code = sympy.latex(self._expression)
            else:
                self._latex_code = str(self._expression)
        except Exception:
            self._latex_code = str(self._expression)
        
        # Llamar al constructor base con el código LaTeX como valor de display
        super().__init__(self._expression, self._latex_code)
    
    def __str__(self):
        """Representación string que muestra el código LaTeX"""
        return self._latex_code
    
    def _sympystr(self, printer):
        """Representación SymPy string"""
        return self._latex_code
    
    def latex_code(self):
        """Retorna el código LaTeX generado"""
        return self._latex_code
    
    def original_expression(self):
        """Retorna la expresión original"""
        return self._expression
    
    def to_image(self, filename=None, dpi=150):
        """
        Convierte el LaTeX a imagen (requiere matplotlib)
        """
        try:
            import matplotlib.pyplot as plt
            import matplotlib
            
            # Configurar matplotlib para usar LaTeX
            matplotlib.rcParams['text.usetex'] = True
            
            fig, ax = plt.subplots(figsize=(10, 2))
            ax.text(0.5, 0.5, f'${self._latex_code}$', 
                   horizontalalignment='center',
                   verticalalignment='center',
                   fontsize=16,
                   transform=ax.transAxes)
            ax.axis('off')
            
            if filename:
                plt.savefig(filename, dpi=dpi, bbox_inches='tight')
                plt.close()
                return f"LaTeX guardado como imagen en: {filename}"
            else:
                plt.show()
                return "LaTeX mostrado en pantalla"
                
        except ImportError:
            return "Error: matplotlib no está instalado. Instalar con: pip install matplotlib"
        except Exception as e:
            return f"Error generando imagen: {e}"

    @staticmethod
    def Helper(input_str):
        """Ayuda contextual para LaTeX"""
        if re.match(r"^\s*[Ll]atex\b", input_str, re.IGNORECASE):
            return '''LaTeX - Conversión de expresiones matemáticas a código LaTeX

Uso: LaTeX[expresion] o latex[expresion]

Ejemplos:
  LaTeX[x**2 + 2*x + 1]    → x^{2} + 2 x + 1
  LaTeX[sin(x)/cos(x)]     → \\frac{\\sin{\\left(x \\right)}}{\\cos{\\left(x \\right)}}
  LaTeX[Integral(x, x)]    → \\int x\\,dx
  LaTeX[sqrt(x**2 + y**2)] → \\sqrt{x^{2} + y^{2}}

Métodos disponibles:
  .latex_code()       - Obtiene el código LaTeX
  .original_expression() - Obtiene la expresión original
  .to_image()         - Convierte a imagen (requiere matplotlib)
  
Nota: Las expresiones pueden ser strings o objetos SymPy'''
        return None

    @staticmethod
    def PopupFunctionList():
        """Lista de métodos sugeridos para autocompletado de LaTeX"""
        return [
            ("latex_code", "Obtiene el código LaTeX generado"),
            ("original_expression", "Obtiene la expresión matemática original"),
            ("to_image", "Convierte el LaTeX a imagen (requiere matplotlib)"),
        ]


def register_classes_in_module():
    """
    Devuelve una lista de clases definidas en este módulo para ser registradas.
    """
    return [
        ("LaTeX", Class_LaTeX, "SympyClassBase", {
            "add_lowercase": True,
            "supports_brackets": True,
            "description": "Conversión de expresiones matemáticas a formato LaTeX"
        }),
    ] 